#include "nBodySim.hpp"
#include <omp.h>
#include <cmath>
#include <string>
#include <chrono>
#include <iostream>
#include <fstream>
#include <iomanip>

nBodySim::nBodySim(std::string datafile) {
    std::string dummy;
    std::ifstream file(datafile);
    file >> dummy >> dummy >> nParticles_;

    for (unsigned i=0; i < nParticles_; ++i) {
        file >> mass_;
    }

    double maxCoord = 0;
    positions_ = new double[3*nParticles_];
    for (unsigned i=0; i < 3; ++i) {
        for (unsigned j=0; j < nParticles_; ++j) {
            file >> positions_[i+3*j];
            if (std::abs(positions_[i+3*j]) > maxCoord) maxCoord = std::abs(positions_[i+3*j]);
        }
    }
    
    velocities_ = new double[3*nParticles_];
    for (unsigned i=0; i < 3; ++i) {
        for (unsigned j=0; j < nParticles_; ++j) {
            file >> velocities_[i+3*j];
        }
    }

    for (unsigned i=0; i < nParticles_; ++i) {
        file >> softening_;
    }
    
    potential_ = new double[nParticles_];
    for (unsigned i=0; i < nParticles_; ++i) {
        file >> potential_[i];
    }

    forces_ = new double[3*nParticles_];
    updateForces();
    
    file.close();

    softening_ = 0.01;

    Vector center{0.0, 0.0, 0.0};
    tree_ = new Tree(mass_, positions_, forces_, softening_, nParticles_, maxCoord, center);

    utils_ = Utilities(nParticles_, positions_, velocities_, forces_, mass_, softening_, tree_);
}

nBodySim::~nBodySim() {
    delete tree_;
    delete[] positions_;
    delete[] velocities_;
    delete[] potential_;
    delete[] forces_;
}

void nBodySim::runSimulation(double dt, unsigned nSteps) {
    const unsigned skip = 10;

    std::ofstream positionsFile("../out/positions.dat");
    std::ofstream energyFile("../out/energy.dat");

    positionsFile << nParticles_ << " " << skip << " " << dt << std::endl;

    for (unsigned i=0; i < nSteps; ++i) {

        write2file(positions_, positionsFile, 3, skip);
        energyFile << utils_.calculateTotalEnergy() << std::endl;

        // Vector COM = utils_.getCenterOfMass();

        // Integrate with leapfrog
        auto start = std::chrono::high_resolution_clock::now();
        kick(dt/2.0);
        drift(dt);
        treeUpdateForces();
        // double dev = utils_.verifyForces();
        kick(dt/2.0);
        updateTree();
        auto stop = std::chrono::high_resolution_clock::now();
        auto duration = std::chrono::duration_cast<std::chrono::milliseconds>(stop - start);

        std::cout << std::right << "Step "    << std::setw(5)  << i
                                // << ": MFM = " << std::setw(12) << utils_.calculateMeanForceMagnitude()
                                // << ", DEV = " << std::setw(12) << dev
                                // << ", MCD = " << std::setw(8)  << utils_.calculateMeanCenterDistance()
                                // << ", MID = " << std::setw(8)  << utils_.calculateMeanInterparticleDistance()
                                // << ", COM = " << "(" << COM[0] << ", " << COM[1] << ", " << COM[2] << ")"
                                // << ", E = "   << std::setw(12) << utils_.calculateTotalEnergy()
                                << ", T = "   << std::setw(8)  << duration.count()/1000.0 << " s"
                                << std::endl;
    }
    positionsFile.close();
}

void nBodySim::updateForces() {
    #pragma omp parallel for
    for (unsigned i=0; i < 3*nParticles_; ++i) {
        forces_[i] = 0.0;
    }
    // loop over each pair of particles and calculate the force between them
    #pragma omp parallel for // schedule(guided)
    for (unsigned i = 0; i < nParticles_; i++) {
        double px = positions_[3*i];
        double py = positions_[3*i+1];
        double pz = positions_[3*i+2];
        double fx = 0.0;
        double fy = 0.0;
        double fz = 0.0;
        for (unsigned j = 0; j < nParticles_; j++) {
            if (j == i) continue;
            // calculate distance between particles
            double dx = positions_[3*j+0] - px;
            double dy = positions_[3*j+1] - py;
            double dz = positions_[3*j+2] - pz;
            double r2 = dx*dx + dy*dy + dz*dz;
            double r = std::sqrt(r2);

            // calculate forces
            double factor = mass_*mass_ / (r2 + softening_*softening_);
            double fxij = factor * dx / r;
            double fyij = factor * dy / r;
            double fzij = factor * dz / r;
            fx += fxij;
            fy += fyij;
            fz += fzij;
            /*
            #pragma omp atomic
            forces_[3*j]   -= fxij;
            #pragma omp atomic
            forces_[3*j+1] -= fyij;
            #pragma omp atomic
            forces_[3*j+2] -= fzij;
            */
        }
        // #pragma omp atomic
        forces_[3*i]   += fx;
        // #pragma omp atomic
        forces_[3*i+1] += fy;
        // #pragma omp atomic
        forces_[3*i+2] += fz;
    }
}

void nBodySim::treeUpdateForces() {
    #pragma omp parallel for
    for (unsigned i=0; i < nParticles_; ++i) {
        tree_->updateForce(i);
    }
}

void nBodySim::kick(double dt) {
    #pragma omp parallel for
    for (unsigned i=0; i < nParticles_; ++i) {
        for (unsigned j=0; j < 3; ++j) {
            velocities_[3*i+j] += forces_[3*i+j] * dt / mass_;
        }
    }
}

void nBodySim::drift(double dt) {
    #pragma omp parallel for
    for (unsigned i=0; i < nParticles_; ++i) {
        for (unsigned j=0; j < 3; ++j) {
            positions_[3*i+j] += velocities_[3*i+j] * dt;
        }
    }
}

void nBodySim::updateTree() {
    tree_->update();
}

void nBodySim::write2file(const double* array, std::ofstream& file, unsigned dim, unsigned skip) const {
    // writes array to file, with N rows and dim columns
    for (unsigned i = 0; i < nParticles_; i+=skip) {
        for (unsigned j = 0; j < dim; j++) {
            file << std::right << std::setw(16) << array[dim*i+j] << " ";
        }
        file << std::endl;
    }
}

void nBodySim::saveState2file(unsigned step, std::ofstream& file) const {
    // save step number
    file << "step: " << step << std::endl;
    // save number of particles
    file << nParticles_ << std::endl;
    // save masses
    for (unsigned i=0; i < nParticles_; ++i) {
        file << mass_ << std::endl;
    }
    // save positions
    for (unsigned i=0; i < 3*nParticles_; ++i) {
        file << positions_[i] << std::endl;
    }
    // save velocities
    for (unsigned i=0; i < 3*nParticles_; ++i) {
        file << velocities_[i] << std::endl;
    }
    // save softening
    for (unsigned i=0; i < nParticles_; ++i) {
        file << softening_ << std::endl;
    }
    // save potential
    for (unsigned i=0; i < nParticles_; ++i) {
        file << potential_[i] << std::endl;
    }
}

void nBodySim::saveTree2file(std::string filename) const {
    std::ofstream file(filename);
    tree_->save2file(file);
}